"""Workflow for creating a new Layer 2 Circuit."""

from typing import Any, Self
from uuid import uuid4

from orchestrator import step, workflow
from orchestrator.forms import FormPage
from orchestrator.targets import Target
from orchestrator.types import SubscriptionLifecycle
from orchestrator.workflow import StepList, begin, done
from orchestrator.workflows.steps import resync, set_status, store_process_subscription
from orchestrator.workflows.utils import wrap_create_initial_input_form
from pydantic import BaseModel, ConfigDict, Field, model_validator
from pydantic_forms.types import FormGenerator, State, UUIDstr
from pydantic_forms.validators import Divider, Label, ReadOnlyField

from gso.products.product_blocks.layer_2_circuit import Layer2CircuitSideBlockInactive, Layer2CircuitType
from gso.products.product_blocks.service_binding_port import ServiceBindingPortInactive
from gso.products.product_types.edge_port import EdgePort
from gso.products.product_types.layer_2_circuit import Layer2Circuit, Layer2CircuitInactive
from gso.services.partners import get_partner_by_name
from gso.utils.helpers import active_edge_port_selector, generate_unique_vc_id, partner_choice
from gso.utils.shared_enums import SBPType
from gso.utils.types.interfaces import BandwidthString
from gso.utils.types.tt_number import TTNumber
from gso.utils.types.virtual_identifiers import VLAN_ID


def initial_input_generator(product_name: str) -> FormGenerator:
    """Gather input from the operator about a new Layer 2 Circuit subscription."""
    geant_partner_id = get_partner_by_name("GEANT")["partner_id"]

    class CreateLayer2CircuitServicePage(FormPage):
        model_config = ConfigDict(title=f"{product_name}")

        tt_number: TTNumber
        partner: partner_choice() = geant_partner_id  # type: ignore[valid-type]
        divider: Divider = Field(None, exclude=True)

        layer_2_circuit_type: Layer2CircuitType
        policer_enabled: bool = False

    initial_user_input = yield CreateLayer2CircuitServicePage

    class Layer2CircuitSideSelection(BaseModel):
        edge_port: active_edge_port_selector(  # type: ignore[valid-type]
            partner_id=initial_user_input.partner if initial_user_input.partner != geant_partner_id else None
        )
        vlan_id: VLAN_ID

    def _vlan_range_field(*, is_tagged: bool) -> VLAN_ID | int:
        """Return the appropriate field type based on whether the circuit is tagged."""
        return VLAN_ID if is_tagged else ReadOnlyField(None, default_type=int)

    class Layer2CircuitServiceSidesPage(FormPage):
        model_config = ConfigDict(title=f"{product_name} - Configure Edge Ports")

        vlan_range_label: Label = Field("Please set a VLAN range, bounds including.", exclude=True)
        vlan_range_lower_bound: _vlan_range_field(  # type: ignore[valid-type]
            is_tagged=initial_user_input.layer_2_circuit_type == Layer2CircuitType.TAGGED
        )
        vlan_range_upper_bound: _vlan_range_field(  # type: ignore[valid-type]
            is_tagged=initial_user_input.layer_2_circuit_type == Layer2CircuitType.TAGGED
        )
        vlan_divider: Divider = Field(None, exclude=True)
        policer_bandwidth: (  # type: ignore[valid-type]
            BandwidthString if initial_user_input.policer_enabled else ReadOnlyField(None, default_type=str)
        )
        geant_sid: str

        layer_2_circuit_side_a: Layer2CircuitSideSelection
        side_divider: Divider = Field(None, exclude=True)
        layer_2_circuit_side_b: Layer2CircuitSideSelection

        @model_validator(mode="after")
        def check_unique_sides(self) -> Self:
            if self.layer_2_circuit_side_a.edge_port == self.layer_2_circuit_side_b.edge_port:
                msg = "Both sides of the circuit cannot be connected to the same edge port"
                raise ValueError(msg)
            return self

    layer_2_circuit_input = yield Layer2CircuitServiceSidesPage

    return {"product_name": product_name} | initial_user_input.model_dump() | layer_2_circuit_input.model_dump()


@step("Create subscription")
def create_subscription(product: UUIDstr, partner: str) -> State:
    """Create a new subscription object in the database."""
    subscription = Layer2CircuitInactive.from_product_id(product, partner)

    return {"subscription": subscription, "subscription_id": subscription.subscription_id}


@step("Initialize subscription")
def initialize_subscription(
    subscription: Layer2CircuitInactive,
    layer_2_circuit_side_a: dict[str, Any],
    layer_2_circuit_side_b: dict[str, Any],
    layer_2_circuit_type: Layer2CircuitType,
    vlan_range_lower_bound: VLAN_ID | None,
    vlan_range_upper_bound: VLAN_ID | None,
    policer_enabled: bool,  # noqa: FBT001
    policer_bandwidth: BandwidthString | None,
    geant_sid: str,
) -> State:
    """Build a subscription object from all user input."""
    layer_2_circuit_sides = []
    for circuit_side_data in [layer_2_circuit_side_a, layer_2_circuit_side_b]:
        sbp = ServiceBindingPortInactive.new(
            uuid4(),
            edge_port=EdgePort.from_subscription(subscription_id=circuit_side_data["edge_port"]).edge_port,
            sbp_type=SBPType.L2,
            vlan_id=circuit_side_data["vlan_id"],
            geant_sid=geant_sid,
            is_tagged=layer_2_circuit_type == Layer2CircuitType.TAGGED,
            custom_firewall_filters=False,
        )
        layer2_circuit_side = Layer2CircuitSideBlockInactive.new(uuid4(), sbp=sbp)
        layer_2_circuit_sides.append(layer2_circuit_side)
    subscription.layer_2_circuit.layer_2_circuit_sides = layer_2_circuit_sides
    subscription.layer_2_circuit.virtual_circuit_id = generate_unique_vc_id()
    subscription.layer_2_circuit.layer_2_circuit_type = layer_2_circuit_type
    subscription.layer_2_circuit.vlan_range_lower_bound = vlan_range_lower_bound
    subscription.layer_2_circuit.vlan_range_upper_bound = vlan_range_upper_bound
    subscription.layer_2_circuit.policer_enabled = policer_enabled
    subscription.layer_2_circuit.bandwidth = policer_bandwidth
    subscription.description = f"{subscription.product.name} - {subscription.layer_2_circuit.virtual_circuit_id}"

    subscription = Layer2Circuit.from_other_lifecycle(subscription, SubscriptionLifecycle.PROVISIONING)

    return {"subscription": subscription}


@workflow(
    "Create Layer 2 Circuit Service",
    initial_input_form=wrap_create_initial_input_form(initial_input_generator),
    target=Target.CREATE,
)
def create_layer_2_circuit() -> StepList:
    """Create a new Layer 2 Circuit service subscription."""
    return (
        begin
        >> create_subscription
        >> store_process_subscription(Target.CREATE)
        >> initialize_subscription
        >> set_status(SubscriptionLifecycle.ACTIVE)
        >> resync
        >> done
    )
